#include <ctype.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <pqxx/connection>
#include <pqxx/cursor.h>
#include <pqxx/transaction>
#include <pqxx/nontransaction>
#include <pqxx/result>

#include <cmath>
#include <math.h>

using namespace PGSTD;
using namespace pqxx;

// ***********************************************************************************

char *usage(void)
{
  return
  "USAGE EXAMPLE: 'pcwsr -p /dev/com1 | pcwsr_to_postgresql'\n"
  "Reads output from 'pcwsr -p' and creates records in a PostgreSQL database.\n"
  "Files required:\n"
  "   pcwsr_to_postgresql.ini - Input.  Must exist with mode 0600 in current dir.\n"
  "   pcwsr_to_postgresql.log - Output. Created in current directory.\n"
  "\n"
  "EXAMPLE of pcwsr_to_postgresql.ini file :-\n"
  "   [DB_Connect_Args]\n"
  "   # Refer to Postgresql's PQconnectdb() API call for syntax...\n"
  "   dbname=weather host=my.db-server.net user=fred password=yada\n"
  "\n"
  "   [hPa_Correction]\n"
  "   # The correction to atmospheric pressure due to the altitude at your\n"
  "   # location, to adjust your readings to match published values referenced\n"
  "   # to mean sea level. The correction value you need is +- 0.11 hPa per metre\n"
  "   # above/below mean sea level.  So for example if your location is at\n"
  "   # 57 metres altitude, we must add 57 * 0.11 = 6.27 hPa to your \n"
  "   # weatherstation readings, so in this case you would specify 6.27.\n"
  "   6.27\n\n"
  "Before running pcwsr_to_postgresql for the first time,\n"
  "you must create an empty Postgres database called 'weather'.\n"
  "Assuming that Postgresql is installed, you can do this like so:-\n"
  "	$ createdb weather\n"
  "	$ psql weather < weather.sql\n"
  "The file 'weather.sql' is part of the pcwsr_to_postgresql distribution.\n"
  "If your Postgres server is a remote server or has certain security options set,\n"
  "you may need to specify additional arguments to createdb.\n"
  "For help please see 'man createdb' or see www.Postgresql.org\n\n";
};

/*  BUGS:
//      1. The code is not elegant, mainly due to TX error-handling and SQL data reduction
//         (we only store duplicate values in the DB every so often).
//        Really we need a C++ class for a sensor instance, with methods to say:
//		- when the data sample must be corrupt, because the rate of change is too high (RF TX error);
//		- when the data has changed between samples, so must be written to the SQL DB.
//
//	2. If INSERTing a record causes an SQL error, the entire process will abort.
//	   Perhaps we should just abort the current record. However, this error
//	   should never happen ;-)
//
//  LICENSE:
//	GPL or BSD, whichever you like best.
//
//  AUTHOR:
//	Martin Johnson - email pcwsr@martinshouse.com
//
//  SEE ALSO:
//    	"weather.sql"
//	- creates the (empty) database :-
//		$ createdb weather
//		$ psql < weather.sql
//          weather.sql includes comments to describe the data formats.
//
//    	"pcwsr.c" by Volker Jahns
//	- Reads data from the unidirectional weatherstation
//	  radio receiver. Decodes the binary data packets into ASCII representation.
//
//  TESTED ON:
// 	PostgresSQL 7.3.5 on FreeBSD 4.9, Intel Celeron processor
//
//- VERSION HISTORY:
//	v0.01 - 01-01-2004 - Martin Johnson.	First cut.
//
// ***********************************************************************************
*/

#define EXIT_STATUS_GOOD 0
#define EXIT_STATUS_BAD  1

// Leave this macro in place, for normal operation
#define LOG(s) { cerr << s ; log << s ; }

// Comment out the next line, unless you want lots of big log files.
// The choice is between logging normal events to stderr only,
// or logging them also to the log file (pcwsr_to_postgresql.log)
#define WANT_LOTS_OF_DEBUG_LOGGING 1
#ifdef  WANT_LOTS_OF_DEBUG_LOGGING
#  define LOG_DEBUG(s) { cerr << s << endl; log << s << endl; }
#else
#  define LOG_DEBUG(s) { cerr << s << endl;                   }
#endif

// Say how often we want to generate a data record if no value has changed.
// Samples come in every 180 seconds (so 20 per hour) from each sensor.
// If nothing has changed, let's store data only once per hour.
//
// TODO: This means that graph plotting routines will need to extrapolate
// the most recent value in their time range.
#define MIN_DB_UPDATE_SECS (60*60*1)

void sanitize(char *buf)
{
	// remove any trailing comments
	char *p = strchr(buf, '#');
	if (p)
		*p = '\0';

	// remove any trailing whitespace
	while(true)
	{
		int l = strlen(buf);
		if (l>0 && isblank(buf[l-1]))
			buf[l-1] = '\0';
		else
		break;
	}
}

bool EqualFloats(float f1, float f2, float delta)   // Does f1==f2, to within +- delta?
{
	if (f1 == f2)
		return true;

	float diff = f1 - f2;
	if (diff < 0)
		diff = f2 - f1;

	return (diff < delta);
};

					
char *UpdateDewPoint_T_H(char *time_stamp, char sensortype, char addr, float temperature, int humidity);
void UpdateDewPoint_P(float pressure);
float calculate_dewpoint(float temperature, int humidity, float pressure);

int main(int argc, char *argv[])
{
  int main_exit_status = EXIT_STATUS_BAD;

  if (argc!=1)
  {
  	cout << usage();
	return main_exit_status;
  };

  cerr << "For help, type './pcwsr_to_postgresql -help'" << endl ;

  ofstream log;
  log.open("pcwsr_to_postgresql.log", ios::out | ios::app);
  if (!log.is_open())
  {
  	cerr << "Could not write to log file pcwsr_to_postgresql.log" << endl;
	return main_exit_status;
  };

  {
  	// print startup message
	time_t t = time(NULL);
	char *szTime = ctime(&t);
	string s = "pcwsr_to_postgresql starting at " + string(szTime) + "\n" ;
	LOG(s);
  };

  float fPressureCorrection_hPa = 0.0f;
  char  szDbConnectString[512]; *szDbConnectString='\0';

  {
	// Check that INI file exists with protection mode 0600.
	struct stat stat_rec;
	int status = stat("pcwsr_to_postgresql.ini", &stat_rec);
	if (status==0 &&   (stat_rec.st_mode & 07777) != 0600)
	{
		LOG("Error! INI file pcwsr_to_postgresql.ini exists, but has wrong protection mode.\n"
			"Please chmod to 0600 and try again.\n");
		log.close();
		return main_exit_status;
	}
	else
	if (status!=0)
	{
		LOG("Error! INI file pcwsr_to_postgresql.ini does not exist, or current directory is not readable\n");
		log.close();
		return main_exit_status;
	};
  };
  {
	// Open INI file for reading.
	// Read config parameters into szDbConnectString and fPressureCorrection_hPa.
	ifstream ini_file;
	ini_file.open("pcwsr_to_postgresql.ini", ios::in);
	if (!ini_file.is_open())
	{
		LOG("Could not read INI file pcwsr_to_postgresql.ini\n");
		log.close();
		return main_exit_status;
	};

	{
		// Parse file contents.  Should be something like:
		//   [DB_Connect_Args]
		//   # comments...
		//   dbname=weather host=my.db-server.net user=fred password=yada
		//
		//   [hPa_Correction]
		//   # comments
		//   6.27

#define DB_CONNECT_ARGS_STR "[DB_Connect_Args]"
#define HPA_CORRECTION_STR  "[hPa_Correction]"
		bool bGot_ConnectArgs   = false;
		bool bGot_hPaCorrection = false;
		bool bCorrectionValid   = false;
		while ( (!ini_file.eof())
		        && ((!bGot_ConnectArgs) || (!bGot_hPaCorrection))
		      )
		{
			char buf[512];

			ini_file.getline(buf, sizeof(buf)-1);
			buf[sizeof(buf)-1] = '\0';
			sanitize(buf);

			if ( (!bGot_ConnectArgs) && (!strcasecmp(DB_CONNECT_ARGS_STR, buf)) )
			{
				// next non-blank line MUST be the connect string
				while ((!ini_file.eof()) && (!bGot_ConnectArgs))
				{
					ini_file.getline(buf, sizeof(buf)-1);
					buf[sizeof(buf)-1] = '\0';
					sanitize(buf);
					if (strlen(buf)>0)
					{
						strcpy(szDbConnectString, buf);
						bGot_ConnectArgs = true;

						if (strchr(buf, '['))
						{
							LOG("** Error ** -- Bad [DB_Connect_Args] string in INI file pcwsr_to_postgresql.ini\n");
							LOG(usage());
							log.close();
							return main_exit_status;
						}
					};
				};
			}
			else
			if ( (!bGot_hPaCorrection) && (!strcasecmp(HPA_CORRECTION_STR, buf)) )
			{
				// next non-blank line MUST be the correction value
				while ((!ini_file.eof()) && (!bGot_hPaCorrection))
				{
					ini_file.getline(buf, sizeof(buf)-1);
					buf[sizeof(buf)-1] = '\0';
					sanitize(buf);
					if (strlen(buf)>0)
					{
						bCorrectionValid = (sscanf(buf, "%f", &fPressureCorrection_hPa) == 1);
						bGot_hPaCorrection = true;
					};
				};
			};
		};
		ini_file.close();

		if (!bGot_ConnectArgs)
		{
			LOG("** ERROR ** -- Could not find DB Connect String in INI file pcwsr_to_postgresql.ini\n");
			LOG(usage());
			log.close();
			return main_exit_status;
		};
		if (! (bGot_hPaCorrection && bCorrectionValid) )
		{
			LOG("** ERROR ** -- Could not find valid hPa correction value in INI file pcwsr_to_postgresql.ini\n");
			LOG(usage());
			log.close();
			return main_exit_status;
		};

		ostringstream oss;
		oss << fPressureCorrection_hPa;
		string s = "Using connect string: '" + string(szDbConnectString) + string("'.\n") +
		string("Using pressure correction value ") + oss.str() + string(" hPa.\n");
		LOG(s);
	};
  };

  try
  {
    // Set up a connection to the RDBMS backend
    LOG("Trying to connect to PostgreSQL...\n");
    connection C(szDbConnectString);

    // Check connection is available, because just declaring a connection doesn't do much...
    // This is just to force an error report early (without waiting for serial port data)
    // if the database isn't really there.
    {
	nontransaction T(C);
	char buf[512];
	sprintf(buf, "SELECT tablename FROM pg_tables WHERE (tablename = 'wind')");
	string s = "SQL: " + string(buf);
	LOG_DEBUG(s);
	result R(T.exec( buf ));	// Will throw exception if database can't be opened.
	LOG_DEBUG("SQL transaction complete");
	if (R.empty())
	{
		// We opened a database, but it probably isn't the weather database
		// because it doesn't contain a table called "wind".
		LOG("ERROR!  Specified Postgres database does not contain an expected table!\n"
		    "Please check the Database connect string in INI file pcwsr_to_postgresql.ini .\n"
		    "If the database exists but is empty (no tables yet) then you must run\n"
		    "	psql weather < weather.sql\n"
		    "to create the tables (assuming your DB is called 'weather').\n"
		    "File weather.sql is in the pcwsr_to_postgresql distribution.\n");
		log.close();
		return main_exit_status;
	};
    };

    LOG("Connected. Waiting for new data...\n");
    // Loop around reading lines from stdin, until we get EOF...
    //	- for each line read:
    //		- parse the thing
    //   	- get the data
    //   	- pick the table we want:
    //		- shove the fields into a new row in the table.
    while (!feof(stdin))
    {
    	char buf[512];
	char *status = fgets(buf, sizeof(buf), stdin);
	if (status==NULL)
	{
		LOG("EOF received, quitting..\n");
		break;	// drop out on EOF
	};
	buf[sizeof(buf)-1] = NULL;	// Ensure any over-size strings are terminated
	if (strlen(buf) > 1)
	{
		buf[strlen(buf)-1] = NULL;	// Assume string ends with \n \0, and change \n to \0
	};
	LOG(buf);LOG("\n");

	// Split line up into comma-delimited fields..
#define MAX_FIELDS 20
	char* pField[MAX_FIELDS];
	for (int i=0; i<MAX_FIELDS; i++)
		pField[i] = NULL;
	int iFields = 0;
	char *p = &buf[0];

	// Split up the input line into at most 20 comma-delimited fields,
	// losing the commas after splitting.
	while (*p != NULL)	// not reached end of string
	{
		if (iFields == MAX_FIELDS)
			break;	// done enough already (should not happen)
		pField[iFields++] = p;	// This field starts at 'p'
		while (*p && *p!=',')
			p++;
		if (*p == ',')
		{
			*p = NULL;
			p++;
		};
	};

	// possible lines to check for in "buf"....
	//'2003-12-31 23:30:39',Wind,0x0f,V,25.8,D,185,+-,67.5
	//'2003-12-31 23:30:49',Rain,0x0f,1407
	//'2003-12-31 23:31:49',Brightness,0x0f,0
	//'2003-12-31 23:32:18',Outdoor,0x08,T,3.5,H,84
	//'2003-12-31 23:32:48',Indoor,0x0f,T,20.9,H,34,P,1004

	if (iFields < 4)		// no real data there
	{
		LOG("Bad data record (less than 4 fields) skipped\n");
		continue;  // skip on to next record.
	};
	char *time_stamp  = pField[0];
	char *sensor_type = pField[1];
	char *sensor_addr = pField[2];

	char addr = sensor_addr[strlen(sensor_addr)-1]; // get the address ascii hex digit
	if (addr=='\0')		// shouldn't happen, but it's illegal in SQL 'char' field
		addr = '0';

	// Now process data depending on what kind of record this is...
	if (!strcmp("Wind", sensor_type))
	{
		// input fields:
		//	'2003-12-31 23:30:39',Wind,0x0f,V,25.8,D,185,+-,67.5
		//	0                     1    2    3 4    5 6   7  8
		//	units				  km/h   deg    deg
		// output fields:
		//    datetime timestamp(0) without time zone NOT NULL,
    		//    speed real NOT NULL,		-- store as m/s
    		//    direction real NOT NULL,
    		//    deviation real NOT NULL
		if (iFields < 9)
		{
			LOG("Wind record: format error, skipped.\n");
			continue;
		};
		float windspeed = 0.0f;
		sscanf(pField[4], "%f", &windspeed);

		// Try to reject bad data caused by radio TX errors. Windspeed is a 14-bit value in km/h * 0.1.
		// Not sure of max possible reading. Sensor reads in km/h * 0.1.  Hardware display max is 200 km/h.
		if (windspeed > 250.0)		// 155 m/h
		{
			char s[300];
			sprintf(s, "Bad wind value? %f km/h - discarded.\n", windspeed);
			LOG(s);
		}
		else
		{
			windspeed = windspeed / 3.6f;	// store as m/s, not km/h

			float direction=0.0f;
			float deviation=0.0f;
			sscanf(pField[6], "%f", &direction);
			sscanf(pField[8], "%f", &deviation);

			// TODO: Try to reject anomolies if rate of change is too high
			// (BUT, we don't know the maximum rate of change)

			// Don't log this data if it's unchanged, unless lots of time has passed
			bool bDoIt = false;
			static time_t last_DB_time = 0L;
			static float last_windspeed = 0.0f;
			static float last_direction = 0.0f;
			static float last_deviation = 0.0f;
			static bool bFirstDBWrite = true;

			if (bFirstDBWrite)
			{
				bFirstDBWrite=false;
				bDoIt = true;
			};

		// Change reverted!
		// Actually this is disastrous for the plotting routines
		// as they plot a line joining the previous value to the new one
		// rather than holding the old value and stepping to the new.
		//	if (windspeed * 3.6f < 0.5f)
		//	{
		//		windspeed = 0.0f;	           // Let's say it's a dead calm...
		//		direction = last_direction;        // And don't bother registering change of wind direction.
		//	};

			if (    (windspeed != last_windspeed)
			     || (direction != last_direction)
			     // || (deviation != last_deviation)	// Let's favour reduced DB size for now, as this field isn't used much
			   )
			{
				bDoIt = true;
			}
			else
			{
				time_t time_now = 0L;
				time(&time_now);
				if (difftime(time_now, last_DB_time) >= MIN_DB_UPDATE_SECS)
					bDoIt = true;
			};

			if (bDoIt)
			{
				time(&last_DB_time);
				last_windspeed = windspeed;
				last_direction = direction;
				last_deviation = deviation;

				nontransaction T(C);
				char buf[512];
				sprintf(buf, "INSERT INTO wind (datetime, speed, direction, deviation) "
					"VALUES (%s, %f, %f, %f);",
					time_stamp, windspeed, direction, deviation);
				string s = "SQL: " + string(buf) ;
				LOG_DEBUG(s);
				T.exec( buf );
				LOG_DEBUG("SQL transaction complete");
			};
		};
	}
	else
	if (!strcmp("Rain", sensor_type))
	{
		// input fields:
		//    	'2004-01-01 02:14:41',Rain,0x0f,1412
		//	0                     1    2    3
		//					^ In impulse counts.
		// output fields:
		//	datetime timestamp without time zone NOT NULL,
		//	rainfall double precision NOT NULL  -- store as metres of rain
		//
		// Need to convert impulses to metres of rain.  0.370 mm per impulse.
		// First time through, produce no data (just read old comulative rain total from DB).
		// Thereafter, compute rainfall since last sample, allowing for wraparound of the 12-bit
		// counter through zero.  Write new DB record with new value.
		//
		if (iFields < 4)
		{
			LOG("Rain record: format error, skipped.\n");
			continue;
		};

		static bool   db_read_attempted = false;
		static double total_rainfall_in_metres = 0.0;	// database value, cumulative total in METRES
		if (!db_read_attempted)
		{
			db_read_attempted = true; // First time through: read the previous total rainfall (in metres) from the database..

			// Execute SQL query to find the most recent record in the "rain" table,
			// and read rain.rainfall into total_rainfall_in_metres ...

			char querybuf[512];
			sprintf(querybuf, "SELECT rainfall FROM rain WHERE datetime = (SELECT max(datetime) FROM rain);" );
			string s = "Reading old rainfall total from database...\nSQL: " + string(querybuf) + string("\n");
			LOG(s);

			work T(C, "rainread");
			result R( T.exec(querybuf) );
			LOG_DEBUG("SQL transaction complete");
			if (R.empty())
			{
				LOG("Empty query result - no rainfall records exist!  This is NORMAL when DB is brand new.\n");
			}
			else
			{
				result::const_iterator c = R.begin();
				if (c != R.end())
				{
					// Get char* pointer to column 0 of this result row: i.e. the rainfall value in metreews
					char *column = (char *) (c[0].c_str());
					string s = "Old DB rainfall value is '" + string(column) + string("'\n");
					LOG(s);

					int count = sscanf(column, "%lf", &total_rainfall_in_metres);
					if (count!=1)
					{
						LOG("sscanf failed on old DB rainfall value!\n");
					}
				}
				else
				{
					LOG("R.end() reached - shouldn't happen\n");
				};
			};
		};

		static bool first_time_through = true;
		static double last_raw_rainfall_in_mm = 0.0;  // last raw value from sensor
		double raw_rainfall_in_mm = 0.0;

#define RAIN_VOLUME_IN_MM_PER_IMPULSE 0.37
		int impulses=0;
		sscanf(pField[3], "%d", &impulses);
		raw_rainfall_in_mm = impulses * RAIN_VOLUME_IN_MM_PER_IMPULSE ;

		if (first_time_through)
		{
			// can't compute new rainfall, as we have nothing to compare it with...
			first_time_through = false;
			last_raw_rainfall_in_mm = raw_rainfall_in_mm;
		}
		else
		{
			double new_rainfall_in_mm = 0.0;
			if (raw_rainfall_in_mm < last_raw_rainfall_in_mm)
			{
				// impulse counter has wrapped through zero, so the actual value
				// for new rainfall is raw_rainfall_in_mm plus (MAX COUNTER VALUE IN MM - last_raw_rainfall_in_mm)
				new_rainfall_in_mm = raw_rainfall_in_mm
					+ (RAIN_VOLUME_IN_MM_PER_IMPULSE * double((unsigned int)0xFFF)) - last_raw_rainfall_in_mm;
			}
			else
			{
				new_rainfall_in_mm = (raw_rainfall_in_mm - last_raw_rainfall_in_mm);
			};

			// We must be especially careful to ignore corrupt data for the rainfall sensor,
			// as the database stores cumulative rainfall, so any error will be permanent!
			static double last_new_rainfall = 0.0;
			static bool   bFirstSample = true;
			if (bFirstSample)
			{
				bFirstSample = false;
				last_new_rainfall = new_rainfall_in_mm;
			};

			// Compare new reading with the reading 3 minutes ago...
			// Only trust this reading if it's same as last time..
			if (new_rainfall_in_mm == last_new_rainfall)
			{
				// Store in the DB only if the value has changed, or if a lot of time has passed since last DB write.
				bool bDoIt = false;
				static time_t last_DB_time = 0L;
				static bool bFirstDBWrite = true;
				if (bFirstDBWrite)
				{
					bFirstDBWrite=false;
					bDoIt = true;			// first time through...
				}
				else if (new_rainfall_in_mm != 0.0)
				{
					bDoIt = true;			//  it has rained...
				}
				else
				{
					time_t time_now = 0L;
					time(&time_now);
					if (difftime(time_now, last_DB_time) >= MIN_DB_UPDATE_SECS)
						bDoIt = true;		// it's been a while..
				};

				if (bDoIt)
				{
					time(&last_DB_time);

					// Now convert new rainfall from millimetres to metres, and add to last value.
					total_rainfall_in_metres += new_rainfall_in_mm / 1000.0 ;
					nontransaction T(C);
					char buf[512];
					sprintf(buf, "INSERT INTO rain (datetime, rainfall) "
						"VALUES (%s, %lf);",
						time_stamp, total_rainfall_in_metres);
					string s = "SQL: " + string(buf) ;
					LOG_DEBUG(s);
					T.exec( buf );
					LOG_DEBUG("SQL transaction complete");

					// Save counter value for next time
					last_raw_rainfall_in_mm = raw_rainfall_in_mm;
				};
			}
			else
			{
				// Remember how much rain we think fell in the last 3 minutes, for the check next time.
				// But limit the rate of increase to 10.0 mm per minute (60cm per hour = highly unlikely)
				// to reduce chance of 2 consecutive RF TX errors causing problems...
				// Hopefully the only drawback should be that continuous/extreme flash floods would cause new data
				// to be delayed until the rain eases for a few minutes.
				if (new_rainfall_in_mm > last_new_rainfall + 30.0)	// 30.0mm over 3 minutes == 1 cm of rain per minute (unlikely)
					last_new_rainfall                 += 30.0;
				else
					last_new_rainfall = new_rainfall_in_mm;		// Normal case, i.e. we'll believe it if we see it twice.
			}
		}; // !first_time_through
	} // !strcmp("Rain", sensor_type)
	else
	if (!strcmp("Brightness", sensor_type))
	{
		// input fields:
		//	'2003-12-31 23:34:27',Brightness,0x0f,0
		//	0		      1          2    3
		//	units				      lx
		// output fields:
		//      datetime timestamp without time zone NOT NULL,
		//      address char NOT NULL,		-- store as hex ascii digit
    		//	lux real NOT NULL
		if (iFields < 4)
		{
			LOG("Brightness record: format error, skipped.\n");
			continue;
		};
		float lux = 0.0f;
		sscanf(pField[3], "%f", &lux);

		if (lux > 2.0e5f)	// 200,000 lux is design max of sensor
		{
			// Should really test rate of change too, but don't know how fast it can change...
			// Hmmm... It could change very rapidly if a bird sits on the sensor!
			char s[300];
			sprintf(s, "Corrupt Brightness record? lux=%f -- sample discarded\n", lux);
			LOG(s);
		}
		else
		{
			// Don't log this data if it's unchanged, unless lots of time has passed
			bool bDoIt = false;
			static time_t last_DB_time = 0L;
			static float last_brightness = 0.0f;
			static bool bFirstDBWrite = true;
			if (bFirstDBWrite)
			{
				bFirstDBWrite=false;
				bDoIt = true;
			}
			else if (lux != last_brightness)
			{
				bDoIt = true;
			}
			else
			{
				time_t time_now = 0L;
				time(&time_now);
				if (difftime(time_now, last_DB_time) >= MIN_DB_UPDATE_SECS)
					bDoIt = true;
			};

			if (bDoIt)
			{
				time(&last_DB_time);
				last_brightness = lux;

				nontransaction T(C);
				char buf[512];
				sprintf(buf, "INSERT INTO brightness (datetime, address, lux) "
					"VALUES (%s, '%c', %f);",
					time_stamp, addr, lux);
				string s = "SQL: " + string(buf);
				LOG_DEBUG(s);
				T.exec( buf );
				LOG_DEBUG("SQL transaction complete");
			};
		};
	}
	else
	if (!strcmp("Outdoor", sensor_type))
	{
		// input fields:
		// 	'2004-01-01 01:09:40',Outdoor,0x08,T,3.0,H,84
		//	0		      1       2    3 4   5 6
		//      units:                               C     %
		// output fields:
		//	datetime timestamp without time zone NOT NULL,
		//      address char NOT NULL,		-- store as hex ascii digit
    		//	temperature real NOT NULL,	-- deg. Centigrade
    		//	humidity int2 NOT NULL		-- 0..100%
		if (iFields < 7)
		{
			LOG("Outdoor record: format error, skipped.\n");
			continue;
		};
		float temperature = 0.0f;
		sscanf(pField[4], "%f", &temperature);
		int humidity = 0;
		sscanf(pField[6], "%d", &humidity);

		// TODO: Implement rate-of-change limiting code to reject data samples
		// if data corruption has occurred due to radio interference.

		// Don't log this data if it's unchanged, unless lots of time has passed.
		// since there are 16 possible one-hex-digit sensor addresses, we need an array for each item...
		bool bDoIt = false;
		static time_t last_DB_time[16] = {0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L };
		static float last_temp[16]     = {0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f };
		static int   last_humidity[16] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
		static bool bFirstDBWrite[16] = {true, true, true, true, true, true, true, true, true, true, true, true, true, true, true, true };

		// Sensor addr should be 0..f. Convert to 0..15
		if (!isxdigit(addr))
		{
			LOG("Bad Outdoor sensor address - ignored\n");
		}
		else
		{
			int address = 0;
			if (isdigit(addr))
				address = addr - '0';
			else
				address = toupper(addr) - 'A' + 10;

			if (bFirstDBWrite[address])
			{
				bFirstDBWrite[address]=false;
				bDoIt = true;
			}
			else if (   (!EqualFloats(temperature, last_temp[address], 0.1))
				  || (humidity    != last_humidity[address])
				)
			{
				bDoIt = true;
			}
			else
			{
				time_t time_now = 0L;
				time(&time_now);
				if (difftime(time_now, last_DB_time[address]) >= MIN_DB_UPDATE_SECS)
					bDoIt = true;
			};

			if (bDoIt)
			{
				time(&last_DB_time[address]);
				last_temp[address]     = temperature;
				last_humidity[address] = humidity;

				nontransaction T(C);
				char buf[512];
				sprintf(buf, "INSERT INTO outdoor (datetime, address, temperature, humidity) "
					"VALUES (%s, '%c', %f, %d);",
					time_stamp, addr, temperature, humidity);
				string s = "SQL: " + string(buf);
				LOG_DEBUG(s);
				T.exec( buf );
				LOG_DEBUG("SQL transaction complete");
			};
			
			char *szDewpoint_SqlBuf = NULL;
			
			szDewpoint_SqlBuf  = UpdateDewPoint_T_H(time_stamp, 'O', addr, temperature, humidity);
			if (szDewpoint_SqlBuf!=NULL)
			{
				nontransaction T(C);
				string s = "SQL: " + string(szDewpoint_SqlBuf);
				LOG_DEBUG(s);
				T.exec( szDewpoint_SqlBuf );
				LOG_DEBUG("SQL transaction complete");
			};
			
		};
	}
	else
	if (!strcmp("Indoor", sensor_type))
	{
		if (iFields < 9)
		{
			LOG("indoor record: format error, skipped.\n");
			continue;
		};

		// input fields:
		//	'2004-01-01 01:12:23',Indoor,0x0f,T,19.2,H,34,P,1000
		//	0                     1      2    3 4    5 6  7 8
		//      units:				    C      %    hPa
		// output fields:
		//	datetime timestamp without time zone NOT NULL,
		//      address char NOT NULL,		-- store as hex ascii digit
    		//	temperature real NOT NULL,	-- deg. Centigrade
    		//	humidity int2 NOT NULL,		-- 0 to 100
    		//	pressure real NOT NULL		-- in Pascals
		//
		// Handy Conversions:
		//	1 millibar = 1 mb = 1 hPa
		//	Pa = [hPa] / 100
		//	1 hecto Pascal = 100 Pa  (hecto = 100)
		//	mmHg = [hPa] * 0.75006
		//	inHg = [HPa] * 0.02953
		//	Absolute air pressure correction = subtract/add 0.11 hPa per metre
		//					   altitude above/below mean sea level
		float temperature = 0.0f;
		sscanf(pField[4], "%f", &temperature);

		int humidity = 0;
		sscanf(pField[6], "%d", &humidity);

		float pressure = 0.0f;
		sscanf(pField[8], "%f", &pressure);

		// Add in the correction factor...
		pressure += fPressureCorrection_hPa ;

		// Convert hPa to Pa...
		pressure /= 100.0f;
		
		UpdateDewPoint_P(pressure);
		
		char* szDewpoint_SqlBuf  = NULL;
		szDewpoint_SqlBuf  = UpdateDewPoint_T_H(time_stamp, 'I', addr, temperature, humidity);
		if (szDewpoint_SqlBuf!=NULL)
		{
			nontransaction T(C);
			string s = "SQL: " + string(szDewpoint_SqlBuf);
			LOG_DEBUG(s);
			T.exec( szDewpoint_SqlBuf );
			LOG_DEBUG("SQL transaction complete");
		};
		

		// TODO: Implement code to limit rate-of-change to eliminate
		// data corruption that might occasionally be caused by radio interference.

		// Don't log this data if it's unchanged, unless lots of time has passed.
		// Warning - This assumes only one indoor sensor is fitted.
		// If this isn't the case, then no data reduction will occur.
		bool bDoIt = false;
		static time_t last_DB_time = 0L;
		static float last_temp = 0.0f;
		static int   last_humidity = 0;
		static float last_pressure = 0.0f;
		static bool bFirstDBWrite = true;
		if (bFirstDBWrite)
		{
			bFirstDBWrite=false;
			bDoIt = true;
		}
		else if (   (!EqualFloats(temperature, last_temp, 0.1))
		          || (humidity != last_humidity)
		          || (!EqualFloats(pressure, last_pressure, 0.001))   // 0.001 Pa = 0.1 hPa
			)
		{
			bDoIt = true;
		}
		else
		{
			time_t time_now = 0L;
			time(&time_now);
			if (difftime(time_now, last_DB_time) >= MIN_DB_UPDATE_SECS)
				bDoIt = true;
		};

		if (bDoIt)
		{
			time(&last_DB_time);
			last_temp = temperature;
			last_humidity = humidity;
			last_pressure = pressure;

			nontransaction T(C);
			char buf[512];
			sprintf(buf, "INSERT INTO indoor (datetime, address, temperature, humidity, pressure) "
				"VALUES (%s, '%c', %f, %d, %f);",
				time_stamp, addr, temperature, humidity, pressure);
			string s = "SQL: " + string(buf);
			LOG_DEBUG(s);
			T.exec( buf );
			LOG_DEBUG("SQL transaction complete");
		};
	}
	else
	{
		string s = "Unexpected sensor type: '" + string(sensor_type) + string("' - record skipped\n");
		LOG(s);
	};

    }; //   while (!feof(stdin))

    main_exit_status = EXIT_STATUS_GOOD;

  } // try
  catch (const sql_error &e)
  {
	// If we're interested in the text of a failed query, we can write separate
	// exception handling code for this type of exception
	string s = "SQL error: "  + string(e.what())
	         + string("\nQuery was: '") + string(e.query()) + string("'\n");
	LOG(s);
  }
  catch (const exception &e)
  {
	// All exceptions thrown by libpqxx are derived from std::exception
	string s = "Exception: " + string(e.what()) + string("\n");
	LOG(s);
  }
  catch (...)
  {
	// This is really unexpected (see above)
	LOG("Unhandled exception \n");
  };

  // print shutdown message
  time_t t = time(NULL);
  char *szTime = ctime(&t);
  string s = "pcwsr_to_postgresql shutting down at " + string(szTime) + string("\n");
  LOG(s);

  log.close();
  return main_exit_status;
}


// Outdoor data for the dew point calculation
static bool  bGotPressure=false;
static float fPressure=0.0f;

// Assume index 0..15 for 'I', 16..31 for 'O' sensors...
static time_t last_dewpoint_time[32] = {0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
					0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L };
static float  last_dewpoint[32]      = {0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f,
 					0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f};
static bool bFirstDewpointWrite[32]  = {true, true, true, true, true, true, true, true, true, true, true, true, true, true, true, true,
					true, true, true, true, true, true, true, true, true, true, true, true, true, true, true, true  };

static char szUpdateDewPoint_T_H[512];
					
char *UpdateDewPoint_T_H(char *time_stamp, char sensortype, char addr, float temperature, int humidity)
{
  // Update dew point data.  If a database write is necessary, return a pointer to a null-terminated string
  // containing the SQL command to be executed.  (This is due to an apparent problem passing Connection objects around in the API I'm using.)
  // If no DB update is needed this time round, then NULL will be returned.
  //
  // Caution: the return value points to static data which must not be changed by the caller.

  // Calculate the Dewpoint based on T + H, using last-known value of P.
  // Log it if it has changed, or if 1 hour has passed since last DB write for this address.
  
  if (!bGotPressure)
  	return NULL;
	
  if (humidity==0)    // soil temp sensor doesn't give this
        return NULL;
	
  float fDewpoint = calculate_dewpoint(temperature, humidity, fPressure);
  bool bDoIt = false;
  
  int address = 0;
  if (isdigit(addr))
	address = addr - '0';
  else
	address = toupper(addr) - 'A' + 10;
    
  // Assume index 0..15 for 'I', 16..31 for 'O' sensors...	
  int addrindex = address;
  if (sensortype!='I')
	addrindex += 16;
	
  if (bFirstDewpointWrite[addrindex])
  {
	bFirstDewpointWrite[addrindex]=false;
	bDoIt = true;
  }
  else if (!EqualFloats(fDewpoint, last_dewpoint[addrindex], 0.15))  // Dewpoint varies a lot based on several factors
								    // so don't flush to disk unless changed by 0.15 deg C
  {
	bDoIt = true;
  }
  else
  {
	time_t time_now = 0L;
	time(&time_now);
	if (difftime(time_now, last_dewpoint_time[addrindex]) >= MIN_DB_UPDATE_SECS)
		bDoIt = true;
  };

  if (bDoIt)
  {
	time(&last_dewpoint_time[addrindex]);
	last_dewpoint[addrindex] = fDewpoint;
  
	sprintf(szUpdateDewPoint_T_H, "INSERT INTO dewpoint (datetime, sensortype, address, dewpointtemp) "
			"VALUES (%s, '%c', '%c', %f);",
			time_stamp, sensortype, addr, fDewpoint);
	
	return szUpdateDewPoint_T_H;
  };
  
  return NULL;
};

void UpdateDewPoint_P(float pressure)
{
  // Update the pressure so that the next call to UpdateDewPoint_T_H() will kick out data if needed.
  fPressure = pressure;
  bGotPressure = true;
};

float calculate_dewpoint(float temperature, int humidity, float pressure)
{
	// Assume temp im deg. C,  rel. humity in %,  pressure in Pa
	// Calculate dewpoint temp in deg. C
	// Source of calc is as follows....
/*
Source: http://www.faqs.org/faqs/meteorology/temp-dewpoint/

> >I'm wondering if anyone could please give me the formula for the
> >calculation of dewpoint temperature given relative humidity, current
> >temperature, and station pressure
> 
> First calculate the saturation vap. pres. es (Pa) at temperature T
> (oC):
> 
> es = 610.78 * exp {A T / (T + B) }  
> 
> where es in Pa, A = 17.2694 and B = 237.3 for T>0 otherwise 265.5.
> Then calculate the actual vapour pressure e (Pa) using 
> 
> e = rH / 100 * es
> 
> where rH is the rel.hum in %. Finally invert the equation for es since
> e = es(Td). The dewpoint temperature Td (oC) is then obtained from
> 
> Td = B  f / { 1 - f  }
> 
> where 
> 
> f = ln ( e / 610.78 ) / A
> 
> (Based on Monteith and Unsworth, 1990, Principles of Environmental
> Physics, sec.ed., Arnold, London, 291pp.  ISBN 0-7131-2931-X. 
> Note however that their equation 2.25 for Td is wrong)
> N.J. Bink
*/
	double A = 17.2694;
	double B = 237.3;
	if (temperature < 0.0f)
		B = 265.5;
	double T = double(temperature);
	
	double es = 610.78 * exp( (A*T) / (T+B) );
	double rH = double(humidity);
	double  e = rH / 100.0 * es;
	
	double f = log ( e / 610.78 ) / A ;
	
	double Td = B * f / ( 1.0 - f ) ;
	
	return (float(Td));
};
