-- *****************************************************************************
--
-- PROGRAM:
--      weather.sql : Creates PostgreSQL database table definitions
--                  for "pcwsr_to_postgresql", a weatherstation
--                  database application.
--
-- USAGE:
--      To create the (empty) database :-
--              $ createdb weather
--              $ psql weather < weather.sql
--      But, you might need extra parameters depending on the setup
--      of your PostgreSQL server.
--
--      The data formats are described in comments below.
-- 
-- LICENSE:
--      GPL or BSD, whichever you like best.
--
-- AUTHOR:
--      Martin Johnson - email pcwsr@martinshouse.com
--
-- SEE ALSO:
--      pcwsr.c
--        - Volker Jahn's C program (radio receiver front-end)
--          which decodes data packets from the radio weatherstation
--          hardware sold by ELV.de etc.
--
--      pcwsr_to_postgresql.cxx
--        - My little C++ program to read from "pcwsr.c",
--          and write the weather data into the PostgreSQL database.
--
-- TESTED ON:
--      PostgresSQL 7.3.5 on FreeBSD 4.9, Intel Celeron processor
--      PostgresSQL 8.1.5 on FreeBSD 6.1, Intel Pentium 4 processor
--
-- VERSION HISTORY:
--      v0.01 - 01-01-2004 - Martin Johnson.    First cut.
--
-- *****************************************************************************
--
-- NOTE:
--      All data in this database is held in standard S.I. units.
--      This means that "pcwsr_to_postgresql" will need to do some conversions
--      from the units reported by the weatherstation hardware.  See comments
--      below each table definition for more info.
--
-- *****************************************************************************

-- Brightness sensor
-- Example of data from sensor:
--      pcwsr:          Wed Dec 31 2003 23:34:27 | brightness (version 12 at address 0x0f) | [klx] 0
--      pcwsr -p:       '2003-12-31 23:34:27',Brightness,0x0f,0

CREATE TABLE brightness (
    datetime timestamp without time zone NOT NULL,
    address character(1) NOT NULL,
    lux real NOT NULL
);
CREATE UNIQUE INDEX brightness_datetime ON brightness USING btree (datetime);

COMMENT ON TABLE brightness IS 'Data from brightness sensor';
COMMENT ON COLUMN brightness.datetime IS 'Date/time (to the second) when the data was recorded. Timezone is UTC.';
COMMENT ON COLUMN brightness.address IS 'Sensor address (last nibble as ascii hex)';
COMMENT ON COLUMN brightness.lux IS 'Brightness in lux (ranges: 0 to 3.12 klx; 3.12 to 25.0 klx; 25.1 to 200 klx).';

-- Indoor sensor
-- Example of data from sensor:
--      pcwsr:          Thu Jan 01 2004 01:12:23 |   indoor (version 12 at address 0x0f) | T[C] 19.2 | H[% rel.hum] 34 | P[hPa] 1000
--      pcwsr -p:               '2004-01-01 01:12:23',Indoor,0x0f,T,19.2,H,34,P,1000
-- 
-- Handy Conversions:
--      1 millibar = 1 mb = 1 hPa
--      Pa = [hPa] / 100
--      1 hecto Pascal = 100 Pa  (hecto = 100)
--      mmHg = [hPa] * 0.75006
--      inHg = [HPa] * 0.02953 
--      Absolute air pressure correction = subtract/add 0.11 hPa per metre altitude above/below mean sea level

CREATE TABLE indoor (
    datetime timestamp without time zone NOT NULL,
    address character(1) NOT NULL,
    temperature real NOT NULL,
    humidity smallint NOT NULL,
    pressure real NOT NULL
);
CREATE UNIQUE INDEX indoor_datetime ON indoor USING btree (address, datetime);
ALTER TABLE indoor CLUSTER ON indoor_datetime;

COMMENT ON TABLE indoor IS 'Data from indoor sensor';
COMMENT ON COLUMN indoor.datetime IS 'Date/time (to the second) when the data was recorded. Timezone is UTC.';
COMMENT ON COLUMN indoor.address IS 'Sensor address (last nibble as ascii hex)';
COMMENT ON COLUMN indoor.temperature IS 'Temperature in degrees Centigrade (+- 0.1 C)';
COMMENT ON COLUMN indoor.humidity IS 'Relative Humidity (0 to 100%)';
COMMENT ON COLUMN indoor.pressure IS 'Atmospheric Pressure in Pascals';

-- Outdoor sensor
-- Example of data from sensor:
--      pcwsr:          Thu Jan 01 2004 01:09:40 |  outdoor (version 12 at address 0x08) | T[C] 3.0 | H[% rel.hum] 84
--      pcwsr -p:               '2004-01-01 01:09:40',Outdoor,0x08,T,3.0,H,84
--      pcwsr:          Thu Jan 01 2004 01:09:40 |  outdoor (version 12 at address 0x08) | T[C] 3.0 | H[% rel.hum] 84
--      pcwsr -p:               '2004-01-01 01:09:40',Outdoor,0x08,T,3.0,H,84

CREATE TABLE outdoor (
    datetime timestamp without time zone NOT NULL,
    address character(1) NOT NULL,
    temperature real NOT NULL,
    humidity smallint NOT NULL
);
CREATE UNIQUE INDEX outdoor_datetime ON outdoor USING btree (address, datetime);
ALTER TABLE outdoor CLUSTER ON outdoor_datetime;

COMMENT ON TABLE outdoor IS 'Data from outdoor sensor';
COMMENT ON COLUMN outdoor.datetime IS 'Date/time (to the second) when the data was recorded. Timezone is UTC.';
COMMENT ON COLUMN outdoor.address IS 'Sensor address (last nibble as ascii hex)';
COMMENT ON COLUMN outdoor.temperature IS 'Temperature in degrees Centigrade (+- 0.1 C)';
COMMENT ON COLUMN outdoor.humidity IS 'Relative Humidity (0 to 100%)';

-- Rainfall sensor
-- Example of data from sensor:
--      pcwsr:          Thu Jan 01 2004 02:14:41 |     rain (version 12 at address 0x0f) | [Impulses] 1412
--      pcwsr -p:               '2004-01-01 02:14:41',Rain,0x0f,1412
-- Input filter program converts impulses to metres of rain, and uses subtraction to produce cumulative total.
--
-- Handy Conversions:
--      1 sensor impulse = 0.300 mm of rain = 0.3 litres per square metre
--      1 inch = [mm] / 25.4
--
-- NOTE: The rain sensor hardware has limited range. It is a 12-bit word that wraps around to zero when full.

CREATE TABLE rain (
    datetime timestamp without time zone NOT NULL,
    rainfall double precision NOT NULL
) ;
CREATE UNIQUE INDEX rain_datetime ON rain USING btree (datetime);

COMMENT ON TABLE rain IS 'Data from rain sensor';
COMMENT ON COLUMN rain.datetime IS 'Date/time (to the second) when the data was recorded. Timezone is UTC.';
COMMENT ON COLUMN rain.rainfall IS 'Cumulative rainfall in metres. Display as mm = litres/square metre';

-- Wind sensor
-- Example of data from sensor:
--      pcwsr:          Thu Jan 01 2004 01:15:28 |     wind (version 12 at address 0x0f) | Speed[km/h] 24.1 | Direction [] 195 +/- 45.0
--      pcwsr -p:               '2004-01-01 01:15:28',Wind,0x0f,V,24.1,D,195,+-,45.0
-- 
-- Handy Conversions:
--      m/s = [km/h] / 3.6


-- Outdoor sensor
-- Example of data from sensor:
--      pcwsr:          Thu Jan 01 2004 01:09:40 |  outdoor (version 12 at address 0x08) | T[C] 3.0 | H[% rel.hum] 84
--      pcwsr -p:               '2004-01-01 01:09:40',Outdoor,0x08,T,3.0,H,84
--      pcwsr:          Thu Jan 01 2004 01:15:28 |     wind (version 12 at address 0x0f) | Speed[km/h] 24.1 | Direction [] 195 +/- 45.0
--      pcwsr -p:               '2004-01-01 01:15:28',Wind,0x0f,V,24.1,D,195,+-,45.0
--
-- Handy Conversions: 
--      m/s = [km/h] / 3.6

CREATE TABLE wind (
    datetime timestamp(0) without time zone NOT NULL,
    speed real NOT NULL,
    direction smallint NOT NULL,
    deviation smallint NOT NULL
);
CREATE UNIQUE INDEX wind_datetime ON wind USING btree (datetime);

COMMENT ON TABLE wind IS 'Data from wind sensor';
COMMENT ON COLUMN wind.datetime IS 'Date/time (to the second) when the data was recorded. Timezone is UTC.';
COMMENT ON COLUMN wind.speed IS 'In metres/second.';
COMMENT ON COLUMN wind.direction IS 'In degrees (0 to 359) from  magnetic north, averaged over the last sample time interval.';
COMMENT ON COLUMN wind.deviation IS 'Deviation (+- X degrees) of wind direction during the last sample time interval.';

-- dewpoint
--       Derived from the other sensors - see formula in C code comments.

CREATE TABLE dewpoint (
    datetime timestamp without time zone NOT NULL,
    sensortype character(1) NOT NULL,
    address character(1) NOT NULL,
    dewpointtemp real NOT NULL
);
CREATE UNIQUE INDEX dewpoint_datetime ON dewpoint USING btree (sensortype, address, datetime);
ALTER TABLE dewpoint CLUSTER ON dewpoint_datetime;

COMMENT ON TABLE dewpoint IS 'Dewpoint temperature calculated from Pressure, Temperature, Relative Humidity';
COMMENT ON COLUMN dewpoint.datetime IS 'Date/time (to the second) when the data was recorded. Timezone is UTC.';
COMMENT ON COLUMN dewpoint.sensortype IS 'Sensor type (O outdoor, I indoor)';
COMMENT ON COLUMN dewpoint.address  IS 'Sensor address (last nibble as ascii hex)';
COMMENT ON COLUMN dewpoint.dewpoint IS 'Dewpoint temperature (degrees Centigrade).';

